#include <pipewire/pipewire.h>
#include <spa/param/audio/format-utils.h>
#include <spa/debug/types.h>
#include <spa/param/audio/type-info.h>
#include <vector>
#include <string>
#include <optional>
#include "../include/pipewire.hpp"

static void on_process(void *userdata)
{
        struct capture_stream *data = static_cast<struct capture_stream *>(userdata);
        struct pw_buffer *b;
        struct spa_buffer *buf;
 
        if ((b = pw_stream_dequeue_buffer(data->stream)) == NULL) {
                pw_log_warn("out of buffers: %m");
                return;
        }
 
        buf = b->buffer;
        if (buf->datas[0].data == NULL)
                return;
 
        //printf("got a frame of size %d\n", buf->datas[0].chunk->size);
 
        pw_stream_queue_buffer(data->stream, b);
}
/* [on_process] */
 
static void on_param_changed(void *userdata, uint32_t id, const struct spa_pod *param)
{
        struct capture_stream *data = static_cast<struct capture_stream *>(userdata);
 
        if (param == NULL || id != SPA_PARAM_Format)
                return;
 
        if (spa_format_parse(param,
                        &data->format.media_type,
                        &data->format.media_subtype) < 0)
                return;
 
        if (data->format.media_type != SPA_MEDIA_TYPE_audio ||
            data->format.media_subtype != SPA_MEDIA_SUBTYPE_raw)
                return;
 
        if (spa_format_audio_raw_parse(param, &data->format.info.raw) < 0)
                return;
 
        printf("got audio format:\n"); 
        printf("    channels: %d\n", data->format.info.raw.channels);
        printf("    rate: %d\n", data->format.info.raw.rate);

}

void register_target_node(struct capture_stream *data, uint32_t id, uint32_t client_id, const char* app_name) {
    struct target_node node = {};
    node.app_name = strdup(app_name);
    node.id = id;
    node.client_id = client_id;

    data->nodes.push_back(node);
}

void register_target_port(struct capture_stream *data, uint32_t id, uint32_t node_id, const char *channel_name) {
    struct target_port port = {};
    port.id = id;
    port.node_id = node_id;
    port.channel_name = strdup(channel_name);

    data->ports.push_back(port);
}

void register_target_input(struct capture_stream *data, uint32_t id) {
    struct target_input port = {
        .id = id,
    };
    data->inputs.push_back(port);
}

bool has_matching_node_or_input(struct capture_stream *capture, uint32_t node_id) {
    // Find the corresponding node.
    bool has_matching_node = false;
    for (auto t : capture->nodes) {
        if (t.id == node_id) {
            has_matching_node = true;
            break;
        }
    }
    if (has_matching_node) {
        return true;
    }

    // Find the corresponding input.
    bool has_matching_input = false;
    for (auto t : capture->inputs) {
        if (t.id == node_id) {
            has_matching_input = true;
            break;
        }
    }
    if (has_matching_input) {
        return true;
    }

    return false;
}

bool connect_port_to_sink(struct capture_stream *data, uint32_t node_id, uint32_t id, const char *channel_name) {
    // Find the corresponding node.
    if (!has_matching_node_or_input(data, node_id)) {
        printf("No matching node found\n");
        return false;
    }

    // Find the correct sink port to attach to.
    uint32_t sink_dst_port_id = 0;
    for (auto sink_port : data->sink_ports) {
        printf("%s = %s\n", sink_port.channel, channel_name);
        if (strcmp(sink_port.channel, channel_name) == 0) {
            sink_dst_port_id = sink_port.id;
            break;
        }
    }
    if (!sink_dst_port_id) {
        return false;
    }

    // Connect the port to the sink.
    struct pw_properties *link_props = pw_properties_new(
        PW_KEY_OBJECT_LINGER, "false",
        NULL
    );
    pw_properties_setf(link_props, PW_KEY_LINK_OUTPUT_NODE, "%u", node_id);
    pw_properties_setf(link_props, PW_KEY_LINK_OUTPUT_PORT, "%u", id);

    pw_properties_setf(link_props, PW_KEY_LINK_INPUT_NODE, "%u", data->sink_id);
    pw_properties_setf(link_props, PW_KEY_LINK_INPUT_PORT, "%u", sink_dst_port_id);

    printf(
        "[DBG] Connecting (%d, %d) -> (%d, %d)\n",
        node_id, id,
        data->sink_id, sink_dst_port_id
    );

    struct pw_proxy *link_proxy = static_cast<struct pw_proxy *>(
        pw_core_create_object(
            data->core, "link-factory",
            PW_TYPE_INTERFACE_Link, PW_VERSION_LINK, &link_props->dict, 0
        )
    );
    data->seq = pw_core_sync(data->core, PW_ID_CORE, data->seq);
    pw_properties_free(link_props);

    if (!link_proxy) {
        printf("[ERR] Failed to connect port %u of node %u to capture sink\n", id, node_id);
        return false;
    }
}

void connect_ports_to_sink(struct capture_stream *data) {
    for (auto port : data->ports) {
        printf("[DBG] Attempting to connect port %d\n", port.id);
        connect_port_to_sink(
            data,
            port.node_id,
            port.id,
            port.channel_name
        );
    }
}

static void registry_event_global(void *raw_data, uint32_t id,
                uint32_t permissions, const char *type, uint32_t version,
                const struct spa_dict *props)
{
    if (!type || !props)
        return;

    struct capture_stream *data = static_cast<struct capture_stream *>(raw_data);

    if (id == data->sink_id) {
        const char *serial = spa_dict_lookup(props, PW_KEY_OBJECT_SERIAL);
        if (!serial) {
            data->sink_serial = 0;
        } else {
            data->sink_serial = strtoul(serial, NULL, 10);
        }
    }

    if (strcmp(type, PW_TYPE_INTERFACE_Port) == 0) {
	    const char *nid, *dir, *chn;
	    if (
            !(nid = spa_dict_lookup(props, PW_KEY_NODE_ID)) ||
            !(dir = spa_dict_lookup(props, PW_KEY_PORT_DIRECTION)) ||
            !(chn = spa_dict_lookup(props, PW_KEY_AUDIO_CHANNEL))
        ) {
	    	return;
	    }

        uint32_t node_id = strtoul(nid, NULL, 10);
        if (strcmp(dir, "in") == 0 && node_id == data->sink_id && data->sink_id != SPA_ID_INVALID) {
            // This port belongs to our own capture sink.
            printf("[DBG] Own sink %d (%s) found\n", id, chn);
            data->sink_ports.push_back(
                { id, strdup(chn), }
            );
        } else if (strcmp(dir, "out") == 0) {
            if (!capture_stream_is_ready(data)) {
                // We're not ready to connect streams, so just track it for later.
                printf("[DBG] Capture sink is not ready yet for %d\n", id);
                register_target_port(
                    data,
                    id,
                    node_id,
                    chn
                );
                return;
            }

            connect_port_to_sink(
                data,
                node_id,
                id,
                chn
            );
        }
    } else if (strcmp(type, PW_TYPE_INTERFACE_Node) == 0) {
		const char *node_name, *media_class;
		if (!(node_name = spa_dict_lookup(props, PW_KEY_NODE_NAME)) ||
			!(media_class = spa_dict_lookup(props, PW_KEY_MEDIA_CLASS))) {
			return;
		}

        if (strcmp(media_class, "Stream/Output/Audio") == 0) {
			const char *node_app_name = spa_dict_lookup(props, PW_KEY_APP_NAME);
			if (!node_app_name) {
				node_app_name = node_name;
			}

            if (data->config.name.has_value()) {
                bool matches = strcmp(
                    node_app_name,
                    data->config.name.value().c_str()
                ) == 0;

                printf("[DBG] Node: name %s matches %d exclude %d\n", node_app_name, matches, data->config.exclude);
                if (!(matches ^ data->config.exclude)) {
                    return;
                }
            }

			uint32_t client_id = 0;
			const char *client_id_str = spa_dict_lookup(props, PW_KEY_CLIENT_ID);
			if (client_id_str) {
				client_id = strtoul(client_id_str, NULL, 10);
			}
            
            register_target_node(
                data,
                id,
                client_id,
                node_app_name
            );
        } else if (strcmp(media_class, "Audio/Source") == 0) {
            if (!data->config.device || data->config.exclude || !data->config.name.has_value()) {
                return;
            }

            if (strcmp(node_name, data->config.name.value().c_str()) != 0) {
                return;
            }

            printf("[DBG] Tracking input source %d (%s)\n", id, node_name);
            register_target_input(data, id);
        }
    }
}

static const struct pw_stream_events stream_events = {
        PW_VERSION_STREAM_EVENTS,
        .param_changed = on_param_changed,
        .process = on_process,
};

static const struct pw_registry_events registry_events = {
    PW_VERSION_REGISTRY_EVENTS,
    .global = registry_event_global,
};

static void on_sink_proxy_bound(void *userdata, uint32_t global_id) {
    struct capture_stream *data = static_cast<struct capture_stream *>(userdata);
    data->sink_id = global_id;
    printf("[DBG] Got proxy sink id %d\n", global_id);
}

static void on_sink_proxy_error(void *data, int seq, int res, const char *message)
{
	printf("[pipewire] App capture sink error: seq:%d res:%d :%s", seq, res, message);
}

static const struct pw_proxy_events sink_proxy_events = {
    PW_VERSION_PROXY_EVENTS,
    .bound = on_sink_proxy_bound,
    .error = on_sink_proxy_error,
};

void init_pipewire() {
    pw_init(NULL, NULL);
}

bool capture_stream_is_ready(struct capture_stream *data) {
    return data->sink_id != SPA_ID_INVALID &&
        data->sink_serial != SPA_ID_INVALID &&
        data->sink_ports.size() == data->config.channels;
}

enum spa_audio_channel *channel_num_to_channels(int num) {
    enum spa_audio_channel channels[8];
    if (num == 1) {
        // Mono
        channels[0] = SPA_AUDIO_CHANNEL_MONO;
        channels[1] = SPA_AUDIO_CHANNEL_UNKNOWN;
        channels[2] = SPA_AUDIO_CHANNEL_UNKNOWN;
        channels[3] = SPA_AUDIO_CHANNEL_UNKNOWN;
        channels[4] = SPA_AUDIO_CHANNEL_UNKNOWN;
        channels[5] = SPA_AUDIO_CHANNEL_UNKNOWN;
        channels[6] = SPA_AUDIO_CHANNEL_UNKNOWN;
        channels[7] = SPA_AUDIO_CHANNEL_UNKNOWN;
    } else if (num == 2) {
        // Probably FL,FR.
        channels[0] = SPA_AUDIO_CHANNEL_FL;
        channels[1] = SPA_AUDIO_CHANNEL_FL;
        channels[2] = SPA_AUDIO_CHANNEL_UNKNOWN;
        channels[3] = SPA_AUDIO_CHANNEL_UNKNOWN;
        channels[4] = SPA_AUDIO_CHANNEL_UNKNOWN;
        channels[5] = SPA_AUDIO_CHANNEL_UNKNOWN;
        channels[6] = SPA_AUDIO_CHANNEL_UNKNOWN;
        channels[7] = SPA_AUDIO_CHANNEL_UNKNOWN;
    }

    return channels;
}

const char *channel_num_to_position(int num) {
    if (num == 1) {
        return "MONO";
    } else if (num == 2) {
        return "FL,FR";
    }
}

struct capture_stream create_capture_stream(struct capture_config config) {
    struct capture_stream data = {
        0,
        sink_id: SPA_ID_INVALID,
        sink_serial: SPA_ID_INVALID,
        seq: 0,
        sink_ports: std::vector<struct sink_port> {},
        nodes: std::vector<struct target_node> {},
        ports: std::vector<struct target_port> {},
        inputs: std::vector<struct target_input> {},
        config: config,
    };
    const struct spa_pod *params[1];
    uint8_t buffer[2048];
    struct spa_pod_builder b = SPA_POD_BUILDER_INIT(buffer, sizeof(buffer));
    struct pw_properties *props;
    pw_init(NULL, NULL);

    data.thread_loop = pw_thread_loop_new("PipeWire thread loop", NULL);
    pw_thread_loop_lock(data.thread_loop);
    if (pw_thread_loop_start(data.thread_loop) < 0) {
        printf("Failed to start thread loop");
        return;
    }

    data.context = pw_context_new(pw_thread_loop_get_loop(data.thread_loop), NULL, 0);
    data.core = pw_context_connect(data.context, NULL, 0);
    pw_core_sync(data.core, PW_ID_CORE, 0);
    //pw_thread_loop_wait(data.thread_loop);
    pw_thread_loop_unlock(data.thread_loop);
    char numbuf[2];
    sprintf(numbuf, "%d", data.config.channels);

    props = pw_properties_new(
        PW_KEY_MEDIA_TYPE, "Audio",
        PW_KEY_MEDIA_CATEGORY, "Capture",
        PW_KEY_MEDIA_ROLE, "Screen",
        PW_KEY_NODE_NAME, "GSR",
        PW_KEY_NODE_VIRTUAL, "true",
        PW_KEY_AUDIO_CHANNELS, numbuf,
        SPA_KEY_AUDIO_POSITION, channel_num_to_position(data.config.channels),
        PW_KEY_FACTORY_NAME, "support.null-audio-sink",
        PW_KEY_MEDIA_CLASS, "Audio/Sink/Internal",
        NULL
    );
    data.sink_proxy = static_cast<pw_proxy *>(
        pw_core_create_object(
            data.core,
            "adapter",
            PW_TYPE_INTERFACE_Node, PW_VERSION_NODE, &props->dict, 0
        )
    );
    pw_proxy_add_listener(
        data.sink_proxy,
        &data.sink_proxy_listener,
        &sink_proxy_events,
        &data
    );
    data.registry = pw_core_get_registry(data.core, PW_VERSION_REGISTRY, 0);
    printf("Got registry\n");
    spa_zero(data.registry_listener);
    pw_registry_add_listener(data.registry, &data.registry_listener, &registry_events, &data);
    printf("Listener registered\n");

    printf("Waiting for id\n");
    while (!capture_stream_is_ready(&data)) {
        printf("Poll\n");
        pw_loop_iterate(pw_thread_loop_get_loop(data.thread_loop), -1);
    }
    printf("Node Setup complete\n");

    // Connect delayed ports to the sink
    //pw_thread_loop_lock(data.thread_loop);
    connect_ports_to_sink(&data);
    //pw_thread_loop_unlock(data.thread_loop);

    auto channels = channel_num_to_channels(data.config.channels);
    params[0] = spa_pod_builder_add_object(
        &b,
        SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat,
        SPA_FORMAT_mediaType, SPA_POD_Id(SPA_MEDIA_TYPE_audio),
        SPA_FORMAT_mediaSubtype, SPA_POD_Id(SPA_MEDIA_SUBTYPE_raw),
        SPA_FORMAT_AUDIO_channels, SPA_POD_Int(data.config.channels),
        SPA_FORMAT_AUDIO_position, SPA_POD_Array(sizeof(enum spa_audio_channel), SPA_TYPE_Id, data.config.channels, channels),
        SPA_FORMAT_AUDIO_format, SPA_POD_CHOICE_ENUM_Id(
            8, SPA_AUDIO_FORMAT_U8, SPA_AUDIO_FORMAT_S16_LE, SPA_AUDIO_FORMAT_S32_LE,
			SPA_AUDIO_FORMAT_F32_LE, SPA_AUDIO_FORMAT_U8P, SPA_AUDIO_FORMAT_S16P,
			SPA_AUDIO_FORMAT_S32P, SPA_AUDIO_FORMAT_F32P
        ) 
    );

    data.stream = pw_stream_new(
        data.core,
        "GSR",
        pw_properties_new(
            PW_KEY_NODE_NAME, "GSR",
            PW_KEY_NODE_DESCRIPTION, "GSR Audio Capture",
            PW_KEY_MEDIA_TYPE, "Audio",
            PW_KEY_MEDIA_CATEGORY, "Capture",
            PW_KEY_MEDIA_ROLE, "Production",
            PW_KEY_NODE_WANT_DRIVER, "true",
            PW_KEY_STREAM_CAPTURE_SINK, "true",
            NULL
        )
    );
    struct pw_properties *stream_props = pw_properties_new(NULL, NULL);
    pw_properties_setf(stream_props, PW_KEY_TARGET_OBJECT, "%u", data.sink_serial);
    pw_stream_update_properties(data.stream, &stream_props->dict);
    pw_properties_free(stream_props);

    pw_stream_connect(
        data.stream,
        PW_DIRECTION_INPUT,
        PW_ID_ANY,
        static_cast<pw_stream_flags>(PW_STREAM_FLAG_AUTOCONNECT | PW_STREAM_FLAG_MAP_BUFFERS),
        params,
        1
    );

    struct spa_hook stream_listener;
    pw_stream_add_listener(
        data.stream,
        &stream_listener,
        &stream_events,
        &data
    );

    while (true) {
        pw_loop_iterate(pw_thread_loop_get_loop(data.thread_loop), -1);
    }
}

void free_capture_stream(struct capture_stream *data) {
    pw_proxy_destroy((struct pw_proxy *) data->registry);
    pw_proxy_destroy(data->sink_proxy);
    pw_stream_destroy(data->stream);
    pw_context_destroy(data->context);
    pw_thread_loop_destroy(data->thread_loop);
}